WHOHAS_CONFIG_VERSION = 2;

WhoHasConfig = {
   version   = WHOHAS_CONFIG_VERSION, -- added in v2
   enabled   = 1,
   totals    = 1,
   stacks    = 1,
   inbox     = 1,
   keyring   = 1,
   bags      = 1,
   equipment = 1,
   vault     = 1, -- added in v2
   allfactions = nil
}

WhoHasData = {
}

WhoHas = {
   savedName = "";
   player = "";
   realm = "";
   faction = "";
   guild = nil;
   tooltipText = {};
   altCache = {};
   playerCache = {};
   vault = {};
   inventoryChanged = 1;
}

-- these are internal strings, not for display
WHOHAS_CATEGORY_INVENTORY = "inventory";
WHOHAS_CATEGORY_BANK      = "bank";
WHOHAS_CATEGORY_INBOX     = "inbox";
WHOHAS_CATEGORY_KEYRING   = "keyring";
WHOHAS_CATEGORY_EQUIPMENT = "equipment";
WHOHAS_CATEGORY_INVBAGS   = "invbags";
WHOHAS_CATEGORY_BANKBAGS  = "bankbags";
WHOHAS_CATEGORY_VAULT     = "vault";
WHOHAS_CATEGORY_TOTAL     = "total";
WHOHAS_CATEGORY_STACK     = "stack";

WhoHas.categories = {
   WHOHAS_CATEGORY_INVENTORY,
   WHOHAS_CATEGORY_BANK,
   WHOHAS_CATEGORY_INBOX,
   WHOHAS_CATEGORY_KEYRING,
   WHOHAS_CATEGORY_EQUIPMENT,
   WHOHAS_CATEGORY_INVBAGS,
   WHOHAS_CATEGORY_BANKBAGS,
}

-------------------------------------------------------------------------------
-- OnLoad
-------------------------------------------------------------------------------

function WhoHas.OnLoad()
   SlashCmdList["WHOHAS"] = WhoHas.ShowConfigFrame;
   tinsert(UISpecialFrames, "WhoHasConfigFrame");

   if (PossessionsData) then
      WhoHas.ScanAlts = WhoHas.ScanAltsPoss;
      WhoHas.ScanPlayer = WhoHas.ScanPlayerPoss;
   elseif (myProfile) then
      WhoHas.ScanAlts = WhoHas.ScanAltsCP;
      WhoHas.ScanPlayer = WhoHas.ScanPlayerCP;
   else
      WhoHas.ScanAlts = WhoHas.DoNothing;
      WhoHas.ScanPlayer = WhoHas.DoNothing;
   end

   WhoHas.HookFunction("ReturnInboxItem", WhoHas.ReturnInboxItem);
   WhoHas.HookFunction("SendMail", WhoHas.SendMail);

   WhoHas.HookMethod(ItemRefTooltip, "SetHyperlink", WhoHas.ShowTooltip);

   WhoHas.HookScript(GameTooltip, "OnTooltipSetItem", WhoHas.OnTooltipSetItem);
   WhoHas.HookScript(GameTooltip, "OnTooltipCleared", WhoHas.OnTooltipCleared);

   if (Possessions_ItemButton_OnEnter) then
      WhoHas_Possessions_ItemButton_OnEnter = Possessions_ItemButton_OnEnter
      Possessions_ItemButton_OnEnter        = WhoHas.Possessions_ItemButton_OnEnter
   end

   for event in pairs(WhoHas.events) do
      this:RegisterEvent(event);
   end
end

-------------------------------------------------------------------------------
-- Utility functions
-------------------------------------------------------------------------------

function WhoHas.ShowConfigFrame()
   WhoHasConfigFrame:Show();
end

function WhoHas.InventoryChanged()
   WhoHas.inventoryChanged = 1;
end

function WhoHas.UpdateGuild()
   local guild = GetGuildInfo("player");

   if (guild ~= WhoHas.guild) then
      WhoHas.guild = guild;

      if (WhoHas.guild and WhoHasData[WhoHas.realm] and WhoHasData[WhoHas.realm][WhoHas.guild]) then
         WhoHas.vault = WhoHasData[WhoHas.realm][WhoHas.guild];
      else
         WhoHas.vault = {};
      end

      WhoHas.savedName = "";
   end
end

function WhoHas.DoNothing()
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------

WhoHas.events = {}

function WhoHas.OnEvent()
   local func = WhoHas.events[event];
   if (func) then
      func(arg1, arg2);
   end
end

function WhoHas.events.ADDON_LOADED(name)
   if (name == "WhoHas") then
      WhoHas.player = UnitName("player");
      WhoHas.realm = GetRealmName();
      WhoHas.faction = UnitFactionGroup("player");
      WhoHas.UpdateGuild();
      WhoHas.ScanAlts();
   elseif (name == "Blizzard_GuildBankUI") then
      WhoHas.HookScript(GuildBankFrame, "OnShow", WhoHas.LoadGuildBank);
      WhoHas.HookFunction("GuildBankFrame_Update", WhoHas.ScanGuildBank);
   end
end

function WhoHas.events.VARIABLES_LOADED()
   -- convert old config format to new
   if (not WhoHasConfig.version or WhoHasConfig.version < 2) then
      WhoHasConfig.vault = 1;
      WhoHasConfig.version = WHOHAS_CONFIG_VERSION;
   end
end

WhoHas.events.PLAYER_GUILD_UPDATE    = WhoHas.UpdateGuild;
WhoHas.events.PLAYER_ENTERING_WORLD  = WhoHas.UpdateGuild;
WhoHas.events.UNIT_INVENTORY_CHANGED = WhoHas.InventoryChanged;
WhoHas.events.BAG_UPDATE             = WhoHas.InventoryChanged;

function WhoHas.OnShow()
   -- OnShow is the backup entry method.  This is only used for
   -- hover tooltips in the world and minimap.
   if (not GameTooltip.WhoHasShowing and not GameTooltip.WhoHasRecipe) then
      WhoHas.ShowTooltip(GameTooltip);
   end
end

function WhoHas.OnHide()
   GameTooltip.WhoHasShowing = nil;
   GameTooltip.WhoHasRecipe = nil;
end

function WhoHas.OnTooltipSetItem()
   -- OnTooltipSetItem is the preferred entry method.  This catches
   -- most tooltips.
   if (not this.WhoHasShowing) then
      -- recipes call OnTooltipSetItem twice for some reason, with
      -- an OnShow in between.  We only want to mod the tooltip
      -- after the second SetItem.
      if (not this.WhoHasRecipe) then
         local name, link = this:GetItem();
         if (link) then
            local _, _, _, _, _, type = GetItemInfo(link);
            if (type == "Recipe") then
               this.WhoHasRecipe = 1;
               return;
            end
         end
      end
      this.WhoHasShowing = 1;
      this.WhoHasRecipe = nil;
      WhoHas.ShowTooltip(this);
   end
end

function WhoHas.OnTooltipCleared()
   this.WhoHasShowing = nil;
end

function WhoHasConfigFrame_OnShow()
end

function WhoHasConfigFrame_OnHide()
   WhoHas.InventoryChanged();
   WhoHas.ScanAlts();
   if (not WhoHasConfig.vault) then
      WhoHas.vault = {};
      WhoHasData = {};
   end
end

-------------------------------------------------------------------------------
-- Hooks
-------------------------------------------------------------------------------

function WhoHas.Possessions_ItemButton_OnEnter(args)
   -- don't doctor tooltips inside of Possessions
   WhoHas.skip = true;
   WhoHas_Possessions_ItemButton_OnEnter(args);
   WhoHas.skip = nil;
end

-------------------------------------------------------------------------------
-- Mail handling
-------------------------------------------------------------------------------

function WhoHas.SendMail(target, subject, body)
   if (WhoHasConfig.inbox) then
      -- proper-case the name
      target = string.upper(string.sub(target, 1, 1)) .. string.lower(string.sub(target, 2));
      if (WhoHas.altCache[target]) then
         for i = 1, 12 do
            local item, _, qty, _ = GetSendMailItem(i);
            if (item) then
               WhoHas.altCache[target][item] = WhoHas.altCache[target][item] or {};
               WhoHas.altCache[target][item][WHOHAS_CATEGORY_INBOX] = (WhoHas.altCache[target][item][WHOHAS_CATEGORY_INBOX] or 0) + qty;
            end
         end
      end
      WhoHas.InventoryChanged()
   end
end

function WhoHas.ReturnInboxItem(mailID)
   if (WhoHasConfig.inbox) then
      local _, _, sender, _ = GetInboxHeaderInfo(mailID);

      if (WhoHas.altCache[sender]) then
         for i = 1, 12 do
            local item, _, qty, _ = GetInboxItem(mailID, i);
            if (item) then
               WhoHas.altCache[sender][item] = WhoHas.altCache[sender][item] or {};
               WhoHas.altCache[sender][item][WHOHAS_CATEGORY_INBOX] = (WhoHas.altCache[sender][item][WHOHAS_CATEGORY_INBOX] or 0) + qty;
            end
         end
      end

      WhoHas.InventoryChanged()
   end
end

-------------------------------------------------------------------------------
-- Hooking functions
-------------------------------------------------------------------------------

function WhoHas.HookMethod(object, name, after, before)
   local func = object[name];
   if (func) then
      object[name] = function(self, ...)
                        if (before) then
                           before(self, ...);
                        end
                        r1, r2, r3, r4, r5, r6 = func(self, ...);
                        if (after) then
                           after(self, ...);
                        end
                        return r1, r2, r3, r4, r5, r6;
                     end
   end
end

function WhoHas.HookFunction(name, after, before)
   local func = _G[name];
   if (func) then
      _G[name] = function(...)
                    if (before) then
                       before(...);
                    end
                    r1, r2, r3, r4, r5, r6 = func(...);
                    if (after) then
                       after(...);
                    end
                    return r1, r2, r3, r4, r5, r6;
                 end
   end
end

function WhoHas.HookScript(object, name, after, before)
   local func = object:GetScript(name);
   if (func) then
      local newfunc = function(...)
                         if (before) then
                            before(...);
                         end
                         r1, r2, r3, r4, r5, r6 = func(...);
                         if (after) then
                            after(...);
                         end
                         return r1, r2, r3, r4, r5, r6;
                      end
      object:SetScript(name, newfunc);
   end
end

-------------------------------------------------------------------------------
-- Tooltip display
-------------------------------------------------------------------------------

function WhoHas.ShowTooltip(tooltip)
   if (tooltip and WhoHasConfig.enabled and not WhoHas.skip) then
      local name = getglobal(tooltip:GetName().."TextLeft1"):GetText();
      if (not name or name == "") then
         return;
      end

      if (WhoHas.inventoryChanged) then
         WhoHas.ScanPlayer();
         WhoHas.inventoryChanged = nil;
         WhoHas.savedName = "";
      end

      if (name ~= WhoHas.savedName) then
         WhoHas.tooltipText = {};
         WhoHas.savedName = name;
         WhoHas.GetText(name, WhoHas.tooltipText);
      end

      for i, line in ipairs(WhoHas.tooltipText) do
         tooltip:AddLine(line);
      end
      tooltip:Show();
   end
end

function WhoHas.GetText(name, text)
   local total = WhoHas.ListOwners(name, text);
   if (WhoHasConfig.totals and total > 0) then
      table.insert(text, string.format(WhoHas.formats[WHOHAS_CATEGORY_TOTAL], total));
   end
   if (WhoHasConfig.stacks) then
      local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, invTexture = GetItemInfo(name);
      if (itemStackCount and itemStackCount > 1) then
         table.insert(text, string.format(WhoHas.formats[WHOHAS_CATEGORY_STACK], itemStackCount));
      end
   end
end

function WhoHas.ListOwners(name, text)
   local total = 0;
   if (WhoHas.playerCache[name]) then
      total = total + WhoHas.ListChar(WhoHas.player, WhoHas.playerCache[name], text);
   end
   for charName, charData in pairs(WhoHas.altCache) do
      if (charData[name]) then
         total = total + WhoHas.ListChar(charName, charData[name], text);
      end
   end
   if (WhoHas.vault[name]) then
      local count = WhoHas.vault[name];
      table.insert(text, string.format(WhoHas.formats[WHOHAS_CATEGORY_VAULT], count));
      total = total + count;
   end
   return total;
end

function WhoHas.ListChar(charName, charData, text)
   local total = 0;
   for i, category in ipairs(WhoHas.categories) do
      local count = charData[category];
      if count and count > 0 then
         table.insert(text, string.format(WhoHas.formats[category], count, charName));
         total = total + count;
      end
   end
   return total;
end

-------------------------------------------------------------------------------
-- Possessions support
-------------------------------------------------------------------------------

WhoHas.Poss = {}
WhoHas.Poss.Lauchsuppe = {
   Inventory = { 0, 1, 2, 3, 4 },
   Bank      = { -1, 5, 6, 7, 8, 9, 10, 11 },
   Keyring   = { -2 },
   Equipment = { -3 },
   Inbox     = { -4 },
   InvBags   = { -5 },
   BankBags  = { -6 }
}

WhoHas.Poss.Siz = {
   Inventory = { 0, 1, 2, 3, 4 },
   Bank      = { -1, 5, 6, 7, 8, 9, 10, 11 },
   Equipment = { -2 },
   Inbox     = { -3 },
   Keyring   = { -4 }
}

function WhoHas.ScanPlayerPoss()
   local slots;

   if (POSS_USEDBANKBAGS_CONTAINER) then
      slots = WhoHas.Poss.Lauchsuppe;
   else
      slots = WhoHas.Poss.Siz;
   end

   local charData;
   if (PossessionsData and PossessionsData[WhoHas.realm]) then
      charData = PossessionsData[WhoHas.realm][string.lower(WhoHas.player)];
   end
   if (charData) then
      WhoHas.playerCache = {};
      WhoHas.ScanCharPoss(WhoHas.player, charData, slots, WhoHas.playerCache);
   end
end

function WhoHas.ScanAltsPoss()
   local slots;

   if (POSS_USEDBANKBAGS_CONTAINER) then
      slots = WhoHas.Poss.Lauchsuppe;
   else
      slots = WhoHas.Poss.Siz;
   end

   if (PossessionsData and PossessionsData[WhoHas.realm]) then
      for charName, charData in pairs(PossessionsData[WhoHas.realm]) do
         if (charName and charData and (WhoHasConfig.allfactions or charData.faction == WhoHas.faction)) then
            -- Possessions lower-cases character names, annoyingly
            charName = string.upper(string.sub(charName, 1, 1)) .. string.sub(charName, 2);
            if (charName ~= WhoHas.player) then
               WhoHas.altCache[charName] = {};
               WhoHas.ScanCharPoss(charName, charData, slots, WhoHas.altCache[charName]);
            end
         end
      end
   end
end

function WhoHas.ScanCharPoss(charName, charData, slots, cache)
   if (charData and charData.items) then
      WhoHas.ScanBagsPoss(charName, WHOHAS_CATEGORY_INVENTORY, charData.items, slots.Inventory, cache);
      WhoHas.ScanBagsPoss(charName, WHOHAS_CATEGORY_BANK, charData.items, slots.Bank, cache);
      if (WhoHasConfig.inbox) then
         WhoHas.ScanBagsPoss(charName, WHOHAS_CATEGORY_INBOX, charData.items, slots.Inbox, cache);
      end
      if (WhoHasConfig.keyring) then
         WhoHas.ScanBagsPoss(charName, WHOHAS_CATEGORY_KEYRING, charData.items, slots.Keyring, cache);
      end
      if (WhoHasConfig.equipment) then
         WhoHas.ScanBagsPoss(charName, WHOHAS_CATEGORY_EQUIPMENT, charData.items, slots.Equipment, cache);
      end
      if (WhoHasConfig.bags and slots.InvBags) then
         WhoHas.ScanBagsPoss(charName, WHOHAS_CATEGORY_INVBAGS, charData.items, slots.InvBags, cache);
      end
      if (WhoHasConfig.bags and slots.BankBags) then
         WhoHas.ScanBagsPoss(charName, WHOHAS_CATEGORY_BANKBAGS, charData.items, slots.BankBags, cache);
      end
   end
end

function WhoHas.ScanBagsPoss(char, slot, bags, bagIndex, cache)
   for _, index in pairs(bagIndex) do
      if (bags[index]) then
         for i, item in pairs(bags[index]) do
            if (item and item[1]) then
               local name = item[1];
               local count = item[3] or 1;
               if (not cache[name]) then
                  cache[name] = {};
               end
               cache[name][slot] = count + (cache[name][slot] or 0);
            end
         end
      end
   end
end

-------------------------------------------------------------------------------
-- CharacterProfiler support
-------------------------------------------------------------------------------

function WhoHas.ScanPlayerCP()
   local charData;
   if (myProfile and myProfile[WhoHas.realm] and myProfile[WhoHas.realm].Character) then
      charData = myProfile[WhoHas.realm].Character[WhoHas.player];
   end
   if (charData) then
      WhoHas.playerCache = {};
      WhoHas.doBagsCP(charName, charData.Inventory, WHOHAS_CATEGORY_INVENTORY, WhoHas.formats.Inventory, WhoHas.playerCache);
      WhoHas.doBagsCP(charName, charData.Bank, WHOHAS_CATEGORY_BANK, WhoHas.formats.Bank, WhoHas.playerCache);
      if (WhoHasConfig.inbox) then
         WhoHas.doInboxCP(charName, charData.MailBox, WHOHAS_CATEGORY_INBOX, WhoHas.formats.Inbox, WhoHas.playerCache);
      end
   end
end

function WhoHas.ScanAltsCP()
   if (myProfile and myProfile[WhoHas.realm] and myProfile[WhoHas.realm].Character) then
      for charName, charData in pairs(myProfile[WhoHas.realm].Character) do
         if (charName ~= WhoHas.player and (WhoHasConfig.allfactions or charData and charData.FactionEn == WhoHas.faction)) then
            WhoHas.altCache[charName] = {};
            WhoHas.doBagsCP(charName, charData.Inventory, WHOHAS_CATEGORY_INVENTORY, WhoHas.formats.Inventory, WhoHas.altCache[charName]);
            WhoHas.doBagsCP(charName, charData.Bank, WHOHAS_CATEGORY_BANK, WhoHas.formats.Bank, WhoHas.altCache[charName]);
            if (WhoHasConfig.inbox) then
               WhoHas.doInboxCP(charName, charData.MailBox, WHOHAS_CATEGORY_INBOX, WhoHas.formats.Inbox, WhoHas.altCache[charName]);
            end
         end
      end
   end
end

function WhoHas.doBagsCP(char, bags, slot, format, cache)
   if (bags) then
      for bag, bagData in pairs(bags) do
         if (bagData.Slots) then
            for i = 1, bagData.Slots do
               local item = bagData.Contents[i]
               if (item and item.Name) then
                  local count = item.Quantity or 1;
                  if (not cache[item.Name]) then
                     cache[item.Name] = {};
                  end
                  cache[item.Name][slot] = count + (cache[item.Name][slot] or 0);
               end
            end
         end
      end
   end
end

function WhoHas.doInboxCP(char, inbox, slot, format, cache)
   if (inbox) then
      for i, msg in ipairs(inbox) do
         if (msg) then
            local item = msg.Item;
            if (item and item.Name) then
               local count = item.Quantity or 1;
               if (not cache[item.Name]) then
                  cache[item.Name] = {};
               end
               cache[item.Name][slot] = count + (cache[item.Name][slot] or 0);
            end
         end
      end
   end
end

-------------------------------------------------------------------------------
-- Guild Bank support
-------------------------------------------------------------------------------

function WhoHas.LoadGuildBank()
   if (WhoHasConfig.vault and not WhoHas.vaultLoaded) then
      WhoHas.guild = GetGuildInfo("player");
      if (WhoHas.guild) then
         -- suppress updates temporarily
         WhoHasConfig.vault = nil;
         for tab = 1, GetNumGuildBankTabs() do
            QueryGuildBankTab(tab);
         end
         -- restore updates and force update now
         WhoHas.vaultLoaded = 1;
         WhoHasConfig.vault = 1;
         WhoHas.ScanGuildBank();
      end
   end
end

function WhoHas.ScanGuildBank()
   if (WhoHasConfig.vault) then
      WhoHas.guild = GetGuildInfo("player");
      if (WhoHas.guild) then
         local cache = {};
         for tab = 1, GetNumGuildBankTabs() do
            for slot = 1, MAX_GUILDBANK_SLOTS_PER_TAB do
               local texture, count, _ = GetGuildBankItemInfo(tab, slot);
               if (texture) then
                  local link = GetGuildBankItemLink(tab, slot);
                  local name = GetItemInfo(link);
                  cache[name] = count + (cache[name] or 0);
               end
            end
         end
         WhoHasData[WhoHas.realm] = WhoHasData[realm] or {};
         WhoHasData[WhoHas.realm][WhoHas.guild] = cache;
         WhoHas.vault = WhoHasData[WhoHas.realm][WhoHas.guild];
      end
      WhoHas.savedName = "";
   end
end
